//Copyright (c) 1997 The Open Group
//
//Permission to use, copy, modify, and distribute this software and its documentation for
//any purpose and without fee is hereby granted, provided that the above copyright notice
//appears in all copies and that both the copyright notice and this permission notice
//appear in supporting documentation, and that the name of ("TOG") or The Open Group not
//be used in advertising or publicity pertaining to distribution of the software without
//specific, written prior permission.
//
//THE OPEN GROUP DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
//IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT
//SHALL TOG BE LIABLE FOR ANY SPECIAL, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
//WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT,
//NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
//PERFORMANCE OF THIS SOFTWARE. 

/* 
 * ORG.opengroup.moabatch.Naked version 0.1
 * by Bill la Forge
 * http://www.opengroup.org/~laforge
 */

package ORG.opengroup.beans;

import java.beans.*;
import java.io.*;

/**
 * Naked is a static class which supports Beans stored in the 
 * directory tree starting at the current working directory.
 * <p>
 * Naked requires that the CLASSPATH begins with ".".
 * Otherwise new beans might not override old beans found elsewhere
 * on the CLASSPATH.
 */
public class Naked
{
    /**
     * Return true if the CLASSPATH begins with ".".
     * Unless this requirement is meet, new beans may not correctly
     * override older beans found elsewhere on the CLASSPATH.
     */
    public static boolean isNaked()
    {
        String classpath=System.getProperty("java.class.path");
        String desired="."+System.getProperty("path.separator");
        String prefix=classpath.substring(0,desired.length());
        return desired.equals(prefix);
    }

    /**
     * Load a Bean. (Uses java.beans.Beans.instantiate.)
     * The name of the bean must be of the form x.y.Z,
     * where x.y is the full package name and Z is the class name.
     * @param name Name of the class or bean
     * @return the instantiated object
     * @exception ClassNotFoundException if the class of the bean could not be found
     * @exception IOException if an I/O error occurs
     */
    public static Object load(String name) 
        throws IOException, ClassNotFoundException
    {
        return Beans.instantiate(null,name);
    }

    /**
     * Create directory from package name.
     * @param pck Package name
     * @return True if the directory exists or was created
     */
    public static boolean mkdirs(String pck)
    {
        String fileName = pck.replace('.',File.separator.charAt(0));
        File file=new File(fileName);
        if (file.exists())
          return true;
        return file.mkdirs();
    }

    /**
     * Save a Bean as a disk file in the directory tree starting
     * at the current working directory.
     * The name of the bean must be of the form x.y.Z,
     * where x.y is the full package name and Z is the class name.
     * The file saved under the name x/y/Z.ser (or x\y\Z.ser).
     * @param obj Object to be saved
     * @param name Name of the object
     * @return True if the object was saved
     * @exception IOException if an I/O error occurs
     */
    public static boolean save(Object obj, String name)
        throws IOException
    {
        String fileName = name.replace('.',File.separator.charAt(0))
                        + ".ser";
        File file=new File(fileName);
        int ndx=name.lastIndexOf(".");
        if (ndx>-1)
            if (!mkdirs(name.substring(0,ndx)))
                return false;
        FileOutputStream fos=new FileOutputStream(file);
        BufferedOutputStream bos=new BufferedOutputStream(fos);
        ObjectOutputStream oos=new ObjectOutputStream(bos);
        oos.writeObject(obj);
        oos.close();
        return true;
    }

    /**
     * Rename a bean. The names must be in the form x.y.Z,
     * where x.y is the full package name and Z is the class name.
     * @param oldNam Name of the bean to be renamed
     * @param newNam The new name of the bean
     * @return True if the rename was successful
     */
    public static boolean rename(String oldNam, String newNam)
    {
        String oldFileName = oldNam.replace('.',File.separator.charAt(0))
                       + ".ser";
        File oldFile=new File(oldFileName);
        if (!oldFile.exists())
            return false;
        String newFileName = newNam.replace('.',File.separator.charAt(0))
                       + ".ser";
        File newFile=new File(newFileName);
        int ndx=newNam.lastIndexOf(".");
        if (ndx>-1)
            if (!mkdirs(newNam.substring(0,ndx)))
                return false;
        return oldFile.renameTo(newFile);
    }

    /**
     * Delete bean. The name must be in the form x.y.Z,
     * where x.y is the full package name and Z is the class name.
     * @param name Name of the bean to be deleted.
     * @return True if the delete was successful.
     */
    public static boolean delete(String name)
    {
        String fileName = name.replace('.',File.separator.charAt(0))
                        + ".ser";
        File file=new File(fileName);
        return file.delete();
    }

    /**
     * Check if bean exists in the naked directory tree.
     * The name must be in the form x.y.Z,
     * where x.y is the full package name and Z is the class name.
     * @param name Name of the bean to be deleted.
     * @return True if the delete was successful.
     */
    public static boolean exists(String name)
    {
        String fileName = name.replace('.',File.separator.charAt(0))
                        + ".ser";
        File file=new File(fileName);
        return file.exists();
    }
}  
