//Copyright (c) 1997 The Open Group
//
//Permission to use, copy, modify, and distribute this software and its documentation for
//any purpose and without fee is hereby granted, provided that the above copyright notice
//appears in all copies and that both the copyright notice and this permission notice
//appear in supporting documentation, and that the name of ("TOG") or The Open Group not
//be used in advertising or publicity pertaining to distribution of the software without
//specific, written prior permission.
//
//THE OPEN GROUP DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
//IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT
//SHALL TOG BE LIABLE FOR ANY SPECIAL, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
//WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT,
//NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
//PERFORMANCE OF THIS SOFTWARE. 

package ORG.opengroup.jar;
import ORG.opengroup.misc.*;
import java.util.zip.*;
import java.util.*;
import java.io.*;
import java.beans.*;
import java.security.*;

/**
 * Holds the contents of a jar file.
 */
public class Bucket implements Cloneable, Serializable
{
	/**
	 * Bucket holding all external classes.
	 */
	transient private Bucket nested=null;

	/**
	 * Assign the bucket in which this bucket is nested.
	 */
	public void setNested(Bucket context)
	{
		nested=context;
	}

	/**
	 * Empty token.
	 */
	private final static String empty="";

	/**
	 * File name of the manifest.
	 */
	protected final static String MANIFEST_NAME="META-INF/MANIFEST.MF";

	/**
	 * The manifest.
	 */
	protected Manifest manifest;

	/**
	 * A serializable set of key/value associations.
	 */
	protected Vector pairs;

	/**
	 * Provides rapid indexing into the pairs vector.
	 */
	transient Hashtable table;

	/**
	 * Loads classes and instantiates objects from this bucket.
	 */
	transient JarClassLoader jcl;

	/**
	 * Returns the ClassLoader
         */
	final public ClassLoader getClassLoader()
	{
		return jcl;
	}

	/**
	 * Associates a key with a value and a set of properties
	 */
	final private static class Entry implements Serializable
	{
		final public String key;
		final public byte value[];
		final public long modified;

		public Entry(String nam, byte v[], long tim)
		{
			key=nam;
			value=v;
			modified=tim;
		}
	}

	/**
	 * Enumerates file names
	 */
	final private class EnumNames implements Enumeration, Serializable
	{
		int index;
		Object name;

		public EnumNames()
		{
			index=0;
			name=null;
		}

		public boolean hasMoreElements()
		{
			synchronized (Bucket.this)
			{
				if (name!=null)
					return true;
				Object e=empty;
				while (pairs.size()>index && empty.equals(e))
					e=pairs.elementAt(index++);
				if (empty.equals(e))
					return false;
				name=((Bucket.Entry)e).key;
				return true;
			}
		}

		public Object nextElement()
			throws NoSuchElementException
		{
			if (name==null && !hasMoreElements())
				throw new NoSuchElementException();
			Object rv=name;
			name=null;
			return rv;
		}
	}

	/**
	 * Loads classes from a Bucket.
	 */
	private class JarClassLoader extends ClassLoader
	{
		private Hashtable classes=new Hashtable();

		public JarClassLoader()
		{
			super();
		}

		public synchronized Class loadClass(String name, boolean resolveIt)
	      	throws ClassNotFoundException 
		{
			Class result=(Class)classes.get(name);
			if (result!=null)
			{
		        	if (resolveIt) 
		            	resolveClass(result);
				return result;
			}
			byte  classData[];
			if (nested!=null)
			{
		        	try 
				{
			            	result = nested.jcl.loadClass(name,false);
			        	if (resolveIt) 
				            	resolveClass(result);
			            	return result;
		        	} catch (ClassNotFoundException e) 
				{
		        	}
			}
	        	/* Check with the primordial class loader */
	        	try 
			{
	            	result = super.findSystemClass(name);
	            	return result;
	        	} catch (ClassNotFoundException e) 
			{
	        	}
	        	/* Try to load it from our repository */
			String fileName=name.replace('.','/')+".class";
	        	classData = get(fileName);
	        	if (classData == null) 
			{
	            	throw new ClassNotFoundException();
	        	}
	        	/* Define it (parse the class file) */
	        	result = defineClass(name,classData, 0, classData.length);
	        	if (result == null) 
			{
	            	throw new ClassFormatError();
	        	}
			classes.put(name,result);
	        	if (resolveIt) 
			{
	            	resolveClass(result);
	        	}
	        	return result;
	    	}
	
		public InputStream getResourceAsStream(String name)
		{
			InputStream is=getSystemResourceAsStream(name);
			if (is!=null)
				return is;
			byte buffer[]=get(name);
			if (buffer==null)
				return null;
			return new ByteArrayInputStream(buffer);
		}
	}
	
	
	/**
	 * Creates a Bucket with no content.
	 */
	public Bucket()
	{
		pairs=new Vector();
		table=null;
		manifest=new Manifest();
	}

	/**
	 * Creates a bucket with the content of a jar file.
	 *
	 * @param 		is			The input stream for the jar file.
	 *
	 * @exception	IOException		May be thrown when
	 *						reading from the input stream.
	 *
	 * @exception	ZipException 	May be thrown when
	 *						processing the contents of the
	 *						jar.
	 */
	public Bucket(InputStream is)
		throws IOException, ZipException
	{
		pairs=new Vector();
		table=null;
		manifest=new Manifest();
		ZipInputStream zis=new ZipInputStream(is);
		ZipEntry ze=zis.getNextEntry();
		while(ze!=null)
		{
			String name=ze.getName();
			if (!ze.isDirectory())
			{
				byte d[]=Reads.read(zis);
				CRC32 crc=new CRC32();
				crc.update(d);
				if (ze.getCrc()!=crc.getValue())
					throw new ZipException("Bad CRC");
				put(name,d,ze.getTime());
			}
			zis.closeEntry();
			ze=zis.getNextEntry();
		}
	}

	/**
	 * Write the bucket in jar file format (compressed)
	 *
	 * @param 		os			The output stream for the jar file.
	 *
	 * @exception	IOException		May be thrown when
	 *						writting to the output stream.
	 *
	 * @exception	ZipException 	May be thrown when
	 *						processing the contents of the
	 *						jar.
	 */
	public synchronized void jar(OutputStream os)
		throws IOException, ZipException
	{
		put(MANIFEST_NAME,manifest.getBytes());
		ZipOutputStream zos=new ZipOutputStream(os);
		int s=pairs.size();
		int i;
		for (i=0; i<s; ++i)
		{
			Object o=pairs.elementAt(i);
			if (!empty.equals(o))
			{
				Entry entry=(Entry)o;
				ZipEntry ze=new ZipEntry(entry.key);
				ze.setTime(entry.modified);
				zos.putNextEntry(ze);
				byte d[]=(byte[])entry.value;
				CRC32 crc=new CRC32();
				crc.update(d);
				ze.setCrc(crc.getValue());
				zos.write(d,0,d.length);
				zos.closeEntry();
			}
		}
		// Original Code
		// zos.flush();

		// Altered by rich_paul@bigfoot.com, to be compatible with
		// java.util.zip.ZipFile, as well as 
		// java.util.zip.ZipInputStream
		zos.finish();
		zos.close();
	}

	/**
	 * Write the bucket in jar file format (uncompressed)
	 *
	 * @param 		os	The output stream for the jar file.
	 *
	 * @exception	IOException	May be thrown when
	 *				writting to the output stream.
	 *
	 * @exception	ZipException 	May be thrown when
	 *				processing the contents of the
	 *				jar.
	 */
	public synchronized void store(OutputStream os)
		throws IOException, ZipException
	{
		put(MANIFEST_NAME,manifest.getBytes());
		ZipOutputStream zos=new ZipOutputStream(os);
		zos.setMethod(ZipOutputStream.STORED);
		int s=pairs.size();
		int i;
		for (i=0; i<s; ++i)
		{
			Object o=pairs.elementAt(i);
			if (!empty.equals(o))
			{
				Entry entry=(Entry)o;
				ZipEntry ze=new ZipEntry(entry.key);
				ze.setTime(entry.modified);
				zos.putNextEntry(ze);
				byte d[]=(byte[])entry.value;
				CRC32 crc=new CRC32();
				crc.update(d);
				ze.setCrc(crc.getValue());
				zos.write(d,0,d.length);
				zos.closeEntry();
			}
		}
		// Original Code
		// zos.flush();

		// Updated based on changes made by rich_paul@bigfoot.com
            // to the jar method.
		zos.finish();
		zos.close();
	}

	/**
	 * Clones the Bucket.
	 */
	public synchronized Object clone()
	{
		Bucket b=new Bucket();
		b.pairs=(Vector)pairs.clone();
		b.manifest=(Manifest)manifest.clone();
		return b;
	}

	/**
	 * Initialize the transient table to provide
	 * fast access by file name.
	 */
	synchronized void restore()
	{
		jcl=new JarClassLoader();
		table=new Hashtable();
		int i;
		int s=pairs.size();
		for (i=0; i<s; ++i)
		{
			Object o=pairs.elementAt(i);
			if (empty.equals(o))
			{
				pairs.setElementAt(empty,i);
			}
			else
			{
				Integer ndx=new Integer(i);
				Entry pair=(Entry)o;
				table.put(pair.key,ndx);
			}
		}
	}

	/**
	 * Either deserialize an existing object or, if there is no such object,
	 * create a new object using the null constructor of the class with the
	 * same name.
	 * @param		name	A dot-seperated name, with no file extension.
	 * @exception	IOException	Error occurred while deserializing.
	 * @exception	ClassNotFoundException	No object or class by that name.
	 */
	public Object getObject(String name)
		throws IOException, ClassNotFoundException
	{
		if (table==null)
			restore();
		return Beans.instantiate(jcl,name);
	}

	/**
	 * Return the byte array associated with the given file name.
	 *
	 * @param 		name	A /-seperated path name, complete with
	 *				file extensions.
	 *
	 * @return			The associated byte array or null.
	 */
	public synchronized byte get(String name)[]
	{
		if (table==null)
			restore();
		if (name.equals(MANIFEST_NAME))
			put(MANIFEST_NAME,manifest.getBytes());
		Integer ndx=(Integer)table.get(name);
		if (ndx==null)
			return null;
		Entry entry=(Entry)pairs.elementAt(ndx.intValue());
		if (entry==null)
			return null;
		return entry.value;
	}

	/**
	 * Return the time when an entry was last modified.
	 *
	 * @param 		name	A /-seperated path name, complete with
	 *				file extensions.
	 *
	 * @return			The time in milliseconds, or 0.
	 */
	public synchronized long modified(String name)
	{
		if (table==null)
			restore();
		Integer ndx=(Integer)table.get(name);
		if (ndx==null)
			return 0;
		Entry entry=(Entry)pairs.elementAt(ndx.intValue());
		if (entry==null)
			return 0;
		return entry.modified;
	}

	/**
	 * Serialize an object and save it as a named resource file in the jar.
	 * This is the exact reverse of java.beans.Beans.instantiate.
	 * Security hashes are included in the manifest.
	 *
	 * @param		name	A dot-seperated name, with no file extension.
	 *				The '.'s are replaced by '/' and ".ser" is appended.
	 *
	 * @param		obj	Object to be saved.
	 *
	 * @exception	IOException	Error occurred while adding the object.
	 *
	 * @see java.beans.Beans#instantiate
	 */
	public synchronized void putObject(String name, Object obj)
		throws IOException
	{
		name=name.replace('.','/')+".ser";
		ByteArrayOutputStream baos=new ByteArrayOutputStream();
		ObjectOutputStream oos=new ObjectOutputStream(baos);
		oos.writeObject(obj);
		oos.flush();
		put(name,baos.toByteArray());
	}

	/**
	 * Associates a name and a value.
	 * Security hashes are included in the manifest.
	 *
	 * @param 		name	A /-seperated path name, complete with
	 *				file extensions.
	 *
	 * @param		value	The associated byte array.
	 */
	public void put(String name, byte value[])
	{
		put(name,value,System.currentTimeMillis());
		if (!name.equals(MANIFEST_NAME))
			manifest.force(name).addHash(value);
	}

	synchronized void put(String name, byte value[], long tim)
	{
		if (table==null)
			restore();
		Entry pair=new Entry(name,value,tim);
		int s=pairs.size();
		int elem;
		Object obj=null;
		int found=-1;
		for (elem=0;elem<s;++elem)
		{
			obj=pairs.elementAt(elem);
			if (empty.equals(obj))
			{
				found=elem;
			}
			else
			{
				Entry p=(Entry)obj;
				if (name.equals(p.key))
				{
					break;
				}
			}
		}
		if (elem==s)
		{
			if (found==-1)
				pairs.addElement(pair);
			else
			{
				elem=found;
				pairs.setElementAt(pair,elem);
			}
		}
		else 
			pairs.setElementAt(pair,elem);
		Integer ndx=new Integer(elem);
		table.put(name,ndx);
		if (name.equals(MANIFEST_NAME))
			manifest=new Manifest(value);
	}

	/**
	 * Return an enumeration of the file names.
	 */
	public Enumeration resources()
	{
		return this.new EnumNames();
	}

	/**
	 * Remove an object and its manifest entry from the bucket.
	 * @param	name		A dot-seperated path,
	 *				with NO file extension.
	 */
	public void removeObject(String name)
	{
		name=name.replace('.','/')+".ser";
		remove(name);
	}

	/**
	 * Remove a file and its manifest entry from the bucket.
	 *
	 * @param 		name	A /-seperated path name, complete with
	 *				file extensions.
	 */
	public synchronized void remove(String name)
	{
		if (table==null)
			restore();
		manifest.remove(name);
		Integer ndx=(Integer)table.get(name);
		if (ndx==null)
			return;
		pairs.setElementAt(empty,ndx.intValue());
		table.remove(name);
	}

	/**
	 * Returns the manifest.
	 */
	public synchronized Manifest getManifest()
	{
		return manifest;
	}

	/**
	 * Validate the jar contents against the manifest security hashes.
	 * Returns true if all are valid.
	 */
	public synchronized boolean validate()
		throws NoSuchAlgorithmException
	{
		Enumeration se=manifest.sections();
		while (se.hasMoreElements())
		{
			ManifestSection ms=(ManifestSection)se.nextElement();
			String n=ms.get("Name");
			byte d[]=get(n);
			if (!ms.validate(d))
				return false;
		}
		return true;
	}

	/**
	 * Determines if two Buckets have common files which differ in their manifests.
	 * This is a very strict test which will detect even differences in the
	 * delimiters used in the manifests. It is this strict to meet
	 * security requirements.
	 * <pr>
	 * Note that for every file in a bucket there is a manifest section which
	 * contains a security hash for that file, so as long as both buckets 
	 * have been validated, one need only compare the manifests.
	 * Returns true if there is a conflict.
	 */
	public synchronized boolean hasConflict(Bucket b)
	{
		//Assumption: b is smaller than this
		Enumeration se=b.getManifest().sections();
		while (se.hasMoreElements())
		{
			ManifestSection ms1=(ManifestSection)se.nextElement();
			String n=ms1.get("Name");
			ManifestSection ms2=manifest.get(n);
			if (ms2!=null)
			{
				String a1[]=ms1.getFull();
				String a2[]=ms2.getFull();
				if (a1.length!=a2.length)
					return true;
				int i;
				for (i=0;i<a1.length;++i)
					if (!a1[i].equals(a2[i]))
						return true;
			}
		}
		return false;
	}

	/**
	 * Update the bucket with the contents of another.
	 */
	public synchronized void update(Bucket b)
	{
		Enumeration se=b.getManifest().sections();
		while (se.hasMoreElements())
		{
			ManifestSection ms=(ManifestSection)se.nextElement();
			String n=ms.get("Name");
			byte d[]=get(n);
			put(n,d,b.modified(n));
			manifest.add(ms);
		}
	}
}  
