//Copyright (c) 1997 The Open Group
//
//Permission to use, copy, modify, and distribute this software and its documentation for
//any purpose and without fee is hereby granted, provided that the above copyright notice
//appears in all copies and that both the copyright notice and this permission notice
//appear in supporting documentation, and that the name of ("TOG") or The Open Group not
//be used in advertising or publicity pertaining to distribution of the software without
//specific, written prior permission.
//
//THE OPEN GROUP DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
//IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT
//SHALL TOG BE LIABLE FOR ANY SPECIAL, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
//WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT,
//NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
//PERFORMANCE OF THIS SOFTWARE. 

/* 
 * ORG.opengroup.moabatch.MoaBatch version 0.1
 * by Bill la Forge
 * http://www.opengroup.org/~laforge
 */

package ORG.opengroup.moabatch;
import ORG.opengroup.beans.*;
import java.beans.*;
import java.io.*;
import java.lang.reflect.*;

/**
 * MoaBatch is a batch oriented (GUIless) property editor. Only properties
 * which can be updated via text can be used by MoaBatch.
 * <p>
 * MoaBatch requires that the CLASSPATH begins with ".;", as serialized
 * objects are created in the directory tree under the working directory.
 * <p>
 * MoaBatch is itself a Bean, which can be run using the Main class.
 *
 * MoaBatch supports indexed properties.
 * <pre>
 *
 * Command list:
 * 
 *     Commands - List commands
 *     AbortOn - Exit on error
 *     AbortOff - No exit on error
 *     Echo X - Display arglist X
 *     EchoOn - Turn on command echo
 *     EchoOff - Turn off command echo
 *     End - End processing of current file
 *     Include X - Process commands in file X
 *     Package X - Create directory from package name X
 *     Delete X - Delete bean X
 *     Rename X Y - Rename bean X to Y
 *     Instantiate X - Creates bean X.ser
 *     Properties X - List the properties of bean X
 *     Limit N - Limits the display of elements of an indexed property
 *     Set X Y Z - Set property Y of bean X to Z
 *     SetAt X Y I Z - Set property Y at I of bean X to Z
 */
public class MoaBatch implements Serializable, Runnable
{
    /**
     * Name of this class.
     */
    static final String MoaBatchClassName="ORG.opengroup.moabatch.MoaBatch";

    /**
     * Error flag.
     */
    boolean fin=false;

    /**
     * Property: EditorPath.
     */
    String editorPath=null;
    public String getEditorPath() {return editorPath;}
    public void setEditorPath(String v) 
    {
        editorPath=v;
    }

    /**
     * Property: InputFile.
     */
    protected String inputFile = "";
    public String getInputFile() {return inputFile;}
    public void setInputFile(String v) {inputFile=v;}

    BufferedReader bufferedReader;

    /**
     * Property: Limit on display of indexed properties.
     */
    protected int limit = 20;
    public int getLimit() {return limit;}
    public void setLimit(int v) {limit=v;}

    /**
     * Property: echo.
     */
    protected boolean echo = false;
    public boolean isEcho() {return echo;}
    public void setEcho(boolean v) {echo=v;}

    /**
     * Property: abort on error.
     */
    protected boolean abort = false;
    public boolean isAbort() {return abort;}
    public void setAbort(boolean v) {abort=v;}

    /**
     * Capture control when constructed or deserialized.
     */
    public MoaBatch()
    {
        if (!Beans.isDesignTime())
	  {
            new Thread(this).start();
        }
    }

    private void writeObject(ObjectOutputStream s)
        throws IOException
    {
        s.defaultWriteObject();
    }

    private void readObject(ObjectInputStream s)
        throws ClassNotFoundException, IOException 
    {
        s.defaultReadObject();
        if (!Beans.isDesignTime())
            new Thread(this).start();
    }


    /**
     * Capture control when constructed or deserialized.
     */
    public void run()
    {
		if (editorPath==null)
		{
			editorPath = EditorPath.getPath();
			System.out.println("default editorPath: "+editorPath);
		}
		else
		{
	            EditorPath.setPath(editorPath);
			System.out.println("editorPath: "+editorPath);
		}
            try
            {
                if (!Naked.isNaked())
                {
                    System.out.println("Error: CLASSPATH must begin with \".\"");
                    return;
                }
    
                //We ask that any instantiated beans not instantiate other beans
                //and not start any tasks.
                Beans.setDesignTime(true);
    
                //MoaBatch provides no gui, everything is done via string editors.
                Beans.setGuiAvailable(false);
    
                //Setup the input stream
                if (inputFile.equals(""))
                    bufferedReader=new BufferedReader(new InputStreamReader(System.in));
                else
                    bufferedReader=new BufferedReader(new FileReader(inputFile));

                //process the input until eof
                String command;
                while (parseCommand(bufferedReader.readLine())
                       || (fin && !abort))
                    fin=false;
            } catch (Exception exception)
            {
                exception.printStackTrace();
            }
    }

    /**
     * Process file
     */
    protected boolean include(String inputFile)
        throws Exception
    {
        BufferedReader bufferedReader=new 
            BufferedReader(new FileReader(inputFile));
        String command;
        while (parseCommand(bufferedReader.readLine())
               || (fin && !abort))
            fin=false;
        return !fin;
    }

    /**
     * Parse command
     */
    protected boolean parseCommand(String text) throws Exception
    {
        if (text==null)
            return false;
        doEcho(text);
        int ndx;
        ndx=text.indexOf(";"); //line delimiter
        if (ndx>-1)
            text=text.substring(0,ndx);
        String command;
        StringBuffer remainder=new StringBuffer();
        command=extract(text,remainder);
        if (command==null)
            return true;
        return evaluate(command,remainder.toString());
    }

    /**
     * Evaluate command
     */
    protected boolean evaluate(String command, String args)
        throws Exception
    {
        if (command.equalsIgnoreCase("abortOn"))
        {
            abort=true;
            return true;
        }
        if (command.equalsIgnoreCase("abortOff"))
        {
            abort=false;
            return true;
        }
        if (command.equalsIgnoreCase("echo"))
            return out(args);
        if (command.equalsIgnoreCase("echoOn"))
        {
            echo=true;
            return true;
        }
        if (command.equalsIgnoreCase("echoOff"))
        {
            echo=false;
            return true;
        }
        if (command.equalsIgnoreCase("commands"))
        {
            commands();
            return true;
        }
        if (command.equalsIgnoreCase("end"))
            return false;
        if (command.equalsIgnoreCase("include"))
            return include(extract(args,new StringBuffer()));
        if (command.equalsIgnoreCase("package"))
            return mkdirs(extract(args,new StringBuffer()));
        if (command.equalsIgnoreCase("delete"))
            return delete(extract(args,new StringBuffer()));
        if (command.equalsIgnoreCase("rename"))
            return rename(args);
        if (command.equalsIgnoreCase("instantiate"))
            return instantiate(extract(args,new StringBuffer()));
        if (command.equalsIgnoreCase("properties"))
            return properties(extract(args,new StringBuffer()));
        if (command.equalsIgnoreCase("limit"))
            return makeLimit(extract(args,new StringBuffer()));
        if (command.equalsIgnoreCase("set"))
            return set(args);
        if (command.equalsIgnoreCase("setat"))
            return setAt(args);
        return error("Unknown command: "+command+" "+args);
    }

    /**
     * Set Limit
     */
    protected boolean makeLimit(String arglist) throws Exception
    {
        Integer i=new Integer(arglist);
        int j=i.intValue();
        if (j<1)
            return error("Limit must be greater than 0");
        setLimit(j);
        return true;
    }

    /**
     * Set bean property
     */
    protected boolean set(String arglist) throws Exception
    {
        StringBuffer rem1=new StringBuffer();
        String args=extract(arglist,rem1);
        if (!validate(args))
            return false;
        StringBuffer rem2=new StringBuffer();
        String prop=extract(rem1.toString(),rem2);
        if (prop==null)
            return error("2nd argument is missing");
        String val=rem2.toString();
        Object obj=Beans.instantiate(null,args);
        PropertyEditorAccess pea=new PropertyEditorAccess(obj);
        if (!pea.selectProperty(prop))
            return error("No editor or not Read/Write");
        String init=pea.getValue();
        if (init==null)
            return error("No text editor");
        pea.setValue(val);
        return save(obj,args);
    }

    /**
     * Set indexed bean property
     */
    protected boolean setAt(String arglist) throws Exception
    {
        StringBuffer rem1=new StringBuffer();
        String args=extract(arglist,rem1);
        if (!validate(args))
            return false;
        StringBuffer rem2=new StringBuffer();
        String prop=extract(rem1.toString(),rem2);
        if (prop==null)
            return error("2nd argument is missing");
        StringBuffer rem3=new StringBuffer();
        String sndx=extract(rem2.toString(),rem3); //index
        if (sndx==null)
            return error("3nd argument is missing");
        Integer indx = new Integer(sndx);
        int jndx = indx.intValue();
        String val=rem3.toString(); //value
        Object obj=Beans.instantiate(null,args);
        PropertyEditorAccess pea=new PropertyEditorAccess(obj);
        if (!pea.selectProperty(prop))
            return error("No editor or not Read/Write");
        String init=pea.getValue(jndx);
        if (init==null)
            return error("No text editor");
        pea.setValue(jndx,val);
        return save(obj,args);
    }

    /**
     * List the beans properties
     */
    protected boolean properties(String args) throws Exception
    {
        if (!validate(args))
            return false;
        Object obj=Beans.instantiate(null,args);
        PropertyEditorAccess pea=new PropertyEditorAccess(obj);
        PropertyDescriptor properties[] = pea.getPropertyDescriptors();
        int i;
        for (i=0;i<properties.length;++i)
        {
            // Don't display hidden or expert properties.
            if (properties[i].isHidden() || properties[i].isExpert()) 
                continue;

            String name = properties[i].getDisplayName();
            if (!pea.selectProperty(i))
                continue;
            String init=pea.getValue();
            if (init!=null)
                out("    "+name+" = \""+init+"\"");
            try 
            {
                int ij;
                init="";
                for (ij=0;ij<limit && init!=null;++ij)
                {
                    init=pea.getValue(ij);
                    if (init!=null)
                        out("    "+name+"["+ij+"] = \""+init+"\"");
                }
                if (init!=null)
                    out("        (Limit reached)");
            } catch (Exception ex) 
            {
                continue;
            }
        }
        return true;
    }

    /**
     * instantiate bean
     */
    protected boolean instantiate(String args) throws Exception
    {
        if (!validate(args))
            return false;
        Object obj=Naked.load(args);
        return save(obj,args);
    }

    protected boolean save(Object obj, String args)
        throws Exception
    {
        if (!validate(args))
            return false;
        if (Naked.save(obj,args))
            return true;
        return error("Unable to save Bean");
    }

    /**
     * rename
     */
    protected boolean rename(String args)
    {
        StringBuffer rem=new StringBuffer();
        String oldNam=extract(args,rem);
        if (!validate(oldNam))
            return false;
        String newNam=extract(rem.toString(),new StringBuffer());
        if (!validate(newNam))
            return false;
        if (Naked.rename(oldNam,newNam))
            return true;
        return error("Unable to rename "+oldNam+" to "+newNam);
    }

    /**
     *  Create directory
     */
    protected boolean mkdirs(String args)
    {
        if (!validate(args))
            return false;
        if (Naked.mkdirs(args))
            return true;
        return error("Unable to create directory");
    }

    /**
     * Delete Bean
     */
    protected boolean delete(String args)
    {
        if (!validate(args))
            return false;
        if (!Naked.exists(args))
            return true;
        if (Naked.delete(args))
            return true;
        return error("Unable to delete Bean");
    }

    /**
     * Check bean or package name
     */
    protected boolean validate(String arg)
    {
        if (arg==null)
            return error("Missing argument");
        if (arg.indexOf("..")>-1)
            return error("Bad name: "+arg);
        if (arg.charAt(0)=='.')
            return error("Bad name: "+arg);
        if (arg.charAt(0)==File.separator.charAt(0))
            return error("Bad name: "+arg);
        return true;
    }

    /**
     * List all commands
     */
    protected void commands()
    {
        out("");
        out("Commands - List commands");
        out("AbortOn - Exit on error");
        out("AbortOff - No exit on error");
        out("Echo X- Display arglist X");
        out("EchoOn - Turn on command echo");
        out("EchoOff - Turn off command echo");
        out("End - End processing of current file");
        out("Include X - Process commands in file X");
        out("Package X - Create directory from package name X");
        out("Delete X - Delete Bean X");
        out("Rename X Y - Rename Bean X to Y");
        out("Instantiate X - Creates bean X.ser");
        out("Properties X - List the properties of bean X");
        out("Limit N - Limits the display of elements of an indexed property");
        out("Set X Y Z - Set property Y of bean X to Z");
        out("SetAt X Y I Z - Set property Y at I of bean X to Z");
        out("");
    }

    /**
     * Extract an argument from a string.
     */
    protected String extract(String text, StringBuffer remainder)
    {
        text=text.trim();
        int ndx;
        int len=text.length();
        if (len==0)
            return null;
        for (ndx=0;ndx<len && text.charAt(ndx)>' ';++ndx);
        if (ndx==len)
            return text;
        remainder.append(text.substring(ndx+1));
        return text.substring(0,ndx);
    }

    void doEcho(String command)
    {
        if (echo)
            out(command);
    }

    /**
     * Output string
     */
    protected boolean out(String txt)
    {
        System.out.println(txt);
        return true;
    }

    /**
     * Output string and raise error flag
     */
    protected boolean error(String txt)
    {
        out(txt);
        fin=true;
        return false;
    }
}
